package mvd.jester.simulator;

import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;
import mvd.jester.model.SystemSetup;
import mvd.jester.priority.RateMonotonic;
import mvd.jester.simulator.internals.ProcessorContext;
import mvd.jester.simulator.internals.SortedTaskContextSet;
import mvd.jester.simulator.internals.TaskContextInterface;


/**
 * AbstractSimulator
 */
public abstract class AbstractSimulator implements SimulatorInterface {

    protected final SystemSetup systemSetup;
    protected final Set<ProcessorContext> processors;
    protected final SortedTaskContextSet readyTasks;
    protected long hyperPeriod;

    AbstractSimulator(SystemSetup systemSetup) {
        this.systemSetup = systemSetup;
        this.readyTasks = new SortedTaskContextSet(new RateMonotonic());
        processors = new HashSet<>();
        for (int i = 0; i < systemSetup.getNumberOfProcessors(); ++i) {
            processors.add(new ProcessorContext(i));
        }
        this.hyperPeriod = systemSetup.getTasks().last().getPeriod() * 2 /* * 10 */;
    }


    protected abstract boolean releaseTasks(long timeStep);

    @Override
    public boolean runSimulation() {
        init();
        for (int t = 0; t < hyperPeriod; ++t) {
            if (!releaseTasks(t)) {
                return false;
            }

            Set<ProcessorContext> sortedProcessors =
                    new TreeSet<>((p1, p2) -> (int) (!p1.getJob().isPresent() ? -1
                            : !p2.getJob().isPresent() ? 1 : -1));

            processors.forEach(p -> sortedProcessors.add(p));
            for (ProcessorContext p : sortedProcessors) {
                for (TaskContextInterface tc : readyTasks) {
                    if (p.acceptTask(tc, t)) {
                        break;
                    }
                }

            }

            for (ProcessorContext p : processors) {
                Optional<TaskContextInterface> optionalTc = p.updateExecution(t);
                if (optionalTc.isPresent()) {
                    TaskContextInterface tc = optionalTc.get();
                    if (t >= tc.getDeadline()) {
                        return false;
                    }
                    readyTasks.remove(optionalTc.get());
                }
            }
        }

        return true;
    }

    private void init() {
        this.readyTasks.clear();
        for (ProcessorContext p : processors) {
            p.setJob(null);
        }
        this.hyperPeriod = systemSetup.getTasks().last().getPeriod() * 2;
    }



}
